import os
from subprocess import CalledProcessError
import unittest

from helpers import (
    DictDiffer,
    command,
    run,
    su,
    unit_info,
    )

from subprocess import CalledProcessError

class TestRun(unittest.TestCase):

    def testSimpleCommand(self):
        # Running a simple command (ls) works and running the command
        # produces a string.
        self.assertIsInstance(run('/bin/ls'), str)

    def testStdoutReturned(self):
        # Running a simple command (ls) works and running the command
        # produces a string.
        self.assertIn('Usage:', run('/bin/ls', '--help'))

    def testSimpleCommand(self):
        # If an error occurs a CalledProcessError is raised with the return
        # code, command executed, and the output of the command.
        with self.assertRaises(CalledProcessError) as info:
            run('ls', '--not a valid switch')
        exception = info.exception
        self.assertEqual(2, exception.returncode)
        self.assertEqual("('ls', '--not a valid switch')", exception.cmd)
        self.assertIn('unrecognized option', exception.output)


class TestCommand(unittest.TestCase):

    def testSimpleCommand(self):
        # Creating a simple command (ls) works and running the command
        # produces a string.
        ls = command('/bin/ls')
        self.assertIsInstance(ls(), str)

    def testArguments(self):
        # Arguments can be passed to commands.
        ls = command('/bin/ls')
        self.assertIn('Usage:', ls('--help'))

    def testMissingExecutable(self):
        # If the command does not exist, an OSError (No such file or
        # directory) is raised.
        bad = command('this command does not exist')
        with self.assertRaises(OSError) as info:
            bad()
        self.assertEqual(2, info.exception.errno)

    def testError(self):
        # If the command returns a non-zero exit code, an exception is raised.
        ls = command('/bin/ls')
        with self.assertRaises(CalledProcessError):
            ls('--not a valid switch')

    def testBakedInArguments(self):
        # Arguments can be passed when creating the command as well as when
        # executing it.
        ll = command('/bin/ls', '-l')
        self.assertIn('rw', ll()) # Assumes a file is r/w in the pwd.
        self.assertIn('Usage:', ll('--help'))

    def testQuoting(self):
        # There is no need to quote special shell characters in commands.
        ls = command('/bin/ls')
        ls('--help', '>')


class TestDictDiffer(unittest.TestCase):

    def testStr(self):
        a = dict(cow='moo', pig='oink')
        b = dict(cow='moo', pig='oinkoink', horse='nay')
        diff = DictDiffer(b, a)
        s = str(diff)
        self.assertIn("added: {'horse': None} -> {'horse': 'nay'}", s)
        self.assertIn("removed: {} -> {}", s)
        self.assertIn("changed: {'pig': 'oink'} -> {'pig': 'oinkoink'}", s)
        self.assertIn("unchanged: ['cow']", s)

    def testStrUnmodified(self):
        a = dict(cow='moo', pig='oink')
        diff = DictDiffer(a, a)
        s = str(diff)
        self.assertEquals('no changes', s)

    def testAddedOrChanged(self):
        a = dict(cow='moo', pig='oink')
        b = dict(cow='moo', pig='oinkoink', horse='nay')
        diff = DictDiffer(b, a)
        expected = set(['horse', 'pig'])
        self.assertEquals(expected, diff.added_or_changed)


class TestUnit_info(unittest.TestCase):

    def make_data(self, state='started'):
        return {
            'machines': {0: {
                'dns-name': 'localhost',
                'instance-id': 'local',
                'instance-state': 'running',
                'state': 'running'}},
            'services': {'foo-service': {
                'charm': 'local:oneiric/foo-service-77',
                'relations': {},
                'units': {'foo-unit/29': {
                    'machine': 0,
                    'public-address': '192.168.122.160',
                    'relations': {},
                    'state': state}}}}}

    def testDataParameter(self):
        # The unit_info function can take a data parameter (primarily for
        # testing) that provides the juju service information to be queried.
        # If not provided the "juju status" command is run and its results
        # parsed.
        unit_info('foo-service', 'state', data=self.make_data())

    def testStateFetching(self):
        # The most common data to fetch about a unit is its state.
        state = unit_info('foo-service', 'state', data=self.make_data())
        self.assertEqual('started', state)

    def testFailedState(self):
        state = unit_info(
            'foo-service', 'state', data=self.make_data(state='bad'))
        self.assertNotEqual('started', state)


current_euid = os.geteuid()
current_egid = os.getegid()
current_home = os.environ['HOME']
example_euid = current_euid + 1
example_egid = current_egid + 1
example_home = '/var/lib/example'
userinfo = {'example_user': dict(
        ids=(example_euid, example_egid), home=example_home)}
effective_values = dict(uid=current_euid, gid=current_egid)


def stub_os_seteuid(value):
    effective_values['uid'] = value


def stub_os_setegid(value):
    effective_values['gid'] = value


class TestSuContextManager(unittest.TestCase):

    def setUp(self):
        import helpers
        self.os_seteuid = os.seteuid
        self.os_setegid = os.setegid
        self.helpers_get_user_ids = helpers.get_user_ids
        self.helpers_get_user_home = helpers.get_user_home
        os.seteuid = stub_os_seteuid
        os.setegid = stub_os_setegid
        helpers.get_user_ids = lambda user: userinfo[user]['ids']
        helpers.get_user_home = lambda user: userinfo[user]['home']

    def tearDown(self):
        import helpers
        os.seteuid = self.os_seteuid
        os.setegid = self.os_setegid
        helpers.get_user_ids = self.helpers_get_user_ids
        helpers.get_user_home = self.helpers_get_user_home

    def testChange(self):
        with su('example_user'):
            self.assertEqual(example_euid, effective_values['uid'])
            self.assertEqual(example_egid, effective_values['gid'])
            self.assertEqual(example_home, os.environ['HOME'])

    def testEnvironment(self):
        with su('example_user') as e:
            self.assertEqual(example_euid, e.uid)
            self.assertEqual(example_egid, e.gid)
            self.assertEqual(example_home, e.home)

    def testRevert(self):
        with su('example_user'):
            pass
        self.assertEqual(current_euid, effective_values['uid'])
        self.assertEqual(current_egid, effective_values['gid'])
        self.assertEqual(current_home, os.environ['HOME'])

    def testRevertAfterFailure(self):
        try:
            with su('example_user'):
                raise RuntimeError()
        except RuntimeError:
            self.assertEqual(current_euid, effective_values['uid'])
            self.assertEqual(current_egid, effective_values['gid'])
            self.assertEqual(current_home, os.environ['HOME'])


if __name__ == '__main__':
    unittest.main()
